/*
    Hotspot Filter for Avisynth 2.5 -- Correct hotspot arising from a
	projector's bulb when shooting projected video. Uses a mask image
	obtained by projecting full white (no film in projector).
	Copyright (C) 2004 A.G.Balakhnin aka Fizick, bag@hotmail.ru, http://bag.hotmail.ru

	The plugin is Avisynth port of Hotspot Filter for VirtualDub
	Copyright (C) 2002 Donald A. Graft

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

	Version 0.1, November 6, 2004 (parameters is compatible with VirtualDub filter)
*/

#include "windows.h"
#include "avisynth.h"
#include "stdio.h"

//-------------------------------------------------------------------------------------------
class HotSpot : public GenericVideoFilter {   
  // HotSpot defines the name of your filter class. 
  // This name is only used internally, and does not affect the name of your filter or similar.
  // This filter extends GenericVideoFilter, which incorporates basic functionality.
  // All functions present in the filter must also be present here.

	//  parameters
	const char * mask;
	int mult;
	int add;

	PClip maskclip;
	int masklast;

public:
  // This defines that these functions are present in your class.
  // These functions must be that same as those actually implemented.
  // Since the functions are "public" they are accessible to other classes.
  // Otherwise they can only be called from functions within the class itself.

	HotSpot(PClip _child, const char * _mask, int _mult, int _add, IScriptEnvironment* env);
  // This is the constructor. It does not return any value, and is always used, 
  //  when an instance of the class is created.
  // Since there is no code in this, this is the definition.

  ~HotSpot();
  // The is the destructor definition. This is called when the filter is destroyed.


	PVideoFrame __stdcall GetFrame(int n, IScriptEnvironment* env);
  // This is the function that AviSynth calls to get a given frame.
  // So when this functions gets called, the filter is supposed to return frame n.
};

// The following is the implementation 
// of the defined functions.

//Here is the acutal constructor code used
HotSpot::HotSpot(PClip _child, const char * _mask, int _mult, int _add,  IScriptEnvironment* env):

	GenericVideoFilter(_child), mask(_mask), mult(_mult), add(_add) {
  // This is the implementation of the constructor.
  // The child clip (source clip) is inherited by the GenericVideoFilter,
  //  where the following variables gets defined:
  //   PClip child;   // Contains the source clip.
  //   VideoInfo vi;  // Contains videoinfo on the source clip.

	if ( mult<0 || mult>1000 )
		env->ThrowError("HotSpot: Mult must be from 0 to 1000 !");
	if ( add<0 || add>200 )
		env->ThrowError("HotSpot: Add must be from 0 to 1000 !");
	
	FILE *maskhand = fopen(mask,"r"); // to check if mask file exist
	if (maskhand == NULL)
		env->ThrowError("HotSpot: no mask file %s", mask);
	else
		fclose(maskhand);

	int start = 0;
	int end = 0; // one frame enough
	AVSValue maskload_args[3] = { mask, start, end }; 
	maskclip= env->Invoke("ImageSource", AVSValue(maskload_args,3)).AsClip();
 
	AVSValue convert_args[1] = { maskclip };
	if (vi.IsYV12()) // make mask type as source type
		maskclip= env->Invoke("ConvertToYV12", AVSValue(convert_args,1)).AsClip();
	else if (vi.IsYUY2())
		maskclip= env->Invoke("ConvertToYUY2", AVSValue(convert_args,1)).AsClip();
	else
		env->ThrowError("HotSpot: video must be YV12 or YUY2");

	AVSValue args_cache[1] = { maskclip }; // add cache
	maskclip = env->Invoke("InternalCache", AVSValue(args_cache,1)).AsClip();


	VideoInfo maskvi = maskclip->GetVideoInfo();

	if (vi.width != maskvi.width || vi.height != maskvi.height )
		env->ThrowError("HotSpot: mask size %d x %d is not as clip size",maskvi.width,maskvi.height);

//	if (vi.pixel_type != maskvi.pixel_type)
//		env->ThrowError("HotSpot: mask type must be as clip type");

	masklast = maskvi.num_frames;

	child->SetCacheHints(CACHE_RANGE,0); // range=0 for spatial filter
	
}
//-------------------------------------------------------------------------------------------

// This is where any actual destructor code used goes
HotSpot::~HotSpot() {
  // This is where you can deallocate any memory you might have used.
}
//-------------------------------------------------------------------------------------------

PVideoFrame __stdcall HotSpot::GetFrame(int n, IScriptEnvironment* env) {
// This is the implementation of the GetFrame function.
// See the header definition for further info.

	float mscale = mult / 40000.0f; // use same factors as VirtualDub plugin HotSpot
	float ascale = add / 100000.0f; 

   // Request frame 'n' from the child (source) clip.
	PVideoFrame src = child->GetFrame(n, env);

	int nmask = 0;//min(n,masklast);// mask frame (currrently use == 0)
	PVideoFrame maskframe = maskclip->GetFrame(nmask, env);

 //If you recieve a frame from PClip->GetFrame(...)
  // you must call env->MakeWritable(&frame) be recieve a valid write pointer.
	env->MakeWritable(&src);

	BYTE * srcp;
	const BYTE * maskp;
	int src_pitch;
	int src_width;
	int src_height;
	int mask_pitch;

	if (vi.IsYV12())
	{
		// This code deals with YV12 colourspace where the Y, U and V information are
		// stored in completely separate memory areas
		srcp = src->GetWritePtr(PLANAR_Y);  // Request a  pointer from the source frame.
		src_pitch = src->GetPitch(PLANAR_Y);
		src_width = src->GetRowSize(PLANAR_Y);
		src_height = src->GetHeight(PLANAR_Y);
		maskp = maskframe->GetReadPtr(PLANAR_Y);
		mask_pitch = maskframe->GetPitch(PLANAR_Y);

		for (int h=0; h<src_height; h++)
		{
			for (int w=0; w<src_width; w++)
			{
				float maskbyte = maskp[w];
				float multf = 1.0f + maskbyte*mscale;
				float addf = maskbyte*maskbyte*ascale;
				int masked = (int)(srcp[w]*multf + addf);
				srcp[w] = min(masked,255);
			}
			srcp += src_pitch;
			maskp += mask_pitch;
		}

	}
	else if (vi.IsYUY2())
	{
		srcp = src->GetWritePtr();  // Request a  pointer from the source frame.
		src_pitch = src->GetPitch();
		src_width = src->GetRowSize();
		src_height = src->GetHeight();
		maskp = maskframe->GetReadPtr();
		mask_pitch = maskframe->GetPitch();

		for (int h=0; h<src_height; h++)
		{
			for (int w=0; w<src_width; w+=2) // step by 2
			{
				float maskbyte = maskp[w];
				float multf = 1.0f + maskbyte*mscale;
				float addf = maskbyte*maskbyte*ascale;
				int masked = (int) (srcp[w]*multf + addf);
				srcp[w] = min(masked,255);
			}
			srcp += src_pitch;
			maskp += mask_pitch;
		}



	}
  // As we now are finished processing the image, we return the destination image.
	return src;
}

//-------------------------------------------------------------------------------------------

// This is the function that created the filter, when the filter has been called.
// This can be used for simple parameter checking, so it is possible to create different filters,
// based on the arguments recieved.

AVSValue __cdecl Create_HotSpot(AVSValue args, void* user_data, IScriptEnvironment* env) {
    return new HotSpot(args[0].AsClip(), // the 0th parameter is the source clip
		 args[1].AsString(), // parameter mask - image file name
		 args[2].AsInt(100), // parameter mult
		 args[3].AsInt(100), // parameter add
		 env);  
    // Calls the constructor with the arguments provied.
}

//-------------------------------------------------------------------------------------------

// The following function is the function that actually registers the filter in AviSynth
// It is called automatically, when the plugin is loaded to see which functions this filter contains.

extern "C" __declspec(dllexport) const char* __stdcall AvisynthPluginInit2(IScriptEnvironment* env) {
    env->AddFunction("HotSpot", "c[mask]s[mult]i[add]i", Create_HotSpot, 0);
    // The AddFunction has the following parameters:
    // AddFunction(Filtername , Arguments, Function to call,0);
    
    // Arguments is a string that defines the types and optional names of the arguments for you filter.
    // c - Video Clip
    // i - Integer number
    // f - Float number
    // s - String
    // b - boolean


    return "`HotSpot' HotSpot plugin";

}

